###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
Simple property package for selective oil permeation
"""

from pyomo.environ import Constraint, Var, Param, NonNegativeReals, Suffix
from pyomo.environ import units

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.base.components import Component
from idaes.core.base.phases import LiquidPhase
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    solve_indexed_blocks,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.exceptions import PropertyPackageError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from watertap.core.util.scaling import transform_property_constraints

# Set up logger
_log = idaeslog.getLogger(__name__)


# this first class sets up the property package and creates a central parameter block
@declare_process_block_class("SopParameterBlock")
class SopParameterData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    def build(self):
        """
        Callable method for Block construction.
        """
        super(SopParameterData, self).build()

        self._state_block_class = SopStateBlock

        # phases
        self.Liq = LiquidPhase()

        # components
        self.H2O = Component()
        self.oil = Component()

        # ---default scaling---
        self.set_default_scaling("temperature", 1e-2)
        self.set_default_scaling("pressure", 1e-6)
        self.set_default_scaling("dens_mass_phase", 1e-3, index="Liq")

        # parameters
        self.visc_d_phase_comp = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={("Liq", "H2O"): 1e-3, ("Liq", "oil"): 3.56e-3},
            units=units.kg * units.m**-1 * units.s**-1,
            doc="Dynamic viscosity",
        )

        self.dens_mass_phase_comp = Param(
            self.phase_list,
            self.component_list,
            mutable=True,
            initialize={("Liq", "H2O"): 1e3, ("Liq", "oil"): 780},
            units=units.kg * units.m**-3,
            doc="Component mass density",
        )

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {
                "flow_mass_phase_comp": {"method": None},
                "temperature": {"method": None},
                "pressure": {"method": None},
                "visc_d_phase_comp": {"method": "_visc_d_phase_comp"},
                "dens_mass_phase_comp": {"method": "_dens_mass_phase_comp"},
                "mass_frac_phase_comp": {"method": "_mass_frac_phase_comp"},
                "flow_vol_phase_comp": {"method": "_flow_vol_phase_comp"},
                "flow_vol_phase": {"method": "_flow_vol_phase"},
                "flow_vol": {"method": "_flow_vol"},
                "flow_mass_phase": {"method": "_flow_mass_phase"},
                "dens_mass_phase": {"method": "_dens_mass_phase"},
                "vol_frac_phase_comp": {"method": "_vol_frac_phase_comp"},
                "conc_mass_phase_comp": {"method": "_conc_mass_phase_comp"},
            }
        )

        obj.add_default_units(
            {
                "time": units.s,
                "length": units.m,
                "mass": units.kg,
                "amount": units.mol,
                "temperature": units.K,
            }
        )


# this second class contains methods that are applied to state blocks as a whole,
# rather than individual elements of indexed state blocks. Essentially it just includes
# the initialization routine
class _SopStateBlock(StateBlock):
    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mass_phase_comp : value at which to initialize
                                               phase component flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature

            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            solver : Solver object to use during initialization if None is provided
                     it will use the default solver for IDAES (default = None)
            optarg : solver options dictionary object (default=None)
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix state variables
        flags = fix_state_vars(self, state_args)
        # Check when the state vars are fixed already result in dof 0
        for k in self.keys():
            dof = degrees_of_freedom(self[k])
            if dof != 0:
                raise PropertyPackageError(
                    "\nWhile initializing {sb_name}, the degrees of freedom "
                    "are {dof}, when zero is required. \nInitialization assumes "
                    "that the state variables should be fixed and that no other "
                    "variables are fixed. \nIf other properties have a "
                    "predetermined value, use the calculate_state method "
                    "before using initialize to determine the values for "
                    "the state variables and avoid fixing the property variables."
                    "".format(sb_name=self.name, dof=dof)
                )

        # ---------------------------------------------------------------------
        skip_solve = True  # skip solve if only state variables are present
        for k in self.keys():
            if number_unfixed_variables(self[k]) != 0:
                skip_solve = False

        if not skip_solve:
            # Initialize properties
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            init_log.info_high(
                "Property initialization: {}.".format(idaeslog.condition(results))
            )

        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        # Unfix state variables
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        revert_state_vars(self, flags)
        init_log.info_high("{} State Released.".format(self.name))


# this third class, provides the instructions to create state blocks
@declare_process_block_class("SopStateBlock", block_class=_SopStateBlock)
class SopStateBlockData(StateBlockData):
    def build(self):
        """Callable method for Block construction."""
        super(SopStateBlockData, self).build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # first we create the state variables
        self.flow_mass_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize={("Liq", "H2O"): 1, ("Liq", "oil"): 0.01},
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.kg / units.s,
            doc="Mass flow rate",
        )

        self.temperature = Var(
            initialize=298.15,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.degK,
            doc="State temperature",
        )

        self.pressure = Var(
            initialize=101325,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.Pa,
            doc="State pressure",
        )

    # -----------------------------------------------------------------------------
    # Property Methods
    # these property methods build variables and constraints on demand (i.e. only when asked
    # for by a user or unit model)
    def _visc_d_phase_comp(self):
        add_object_reference(self, "visc_d_phase_comp", self.params.visc_d_phase_comp)

    def _dens_mass_phase_comp(self):
        add_object_reference(
            self, "dens_mass_phase_comp", self.params.dens_mass_phase_comp
        )

    def _mass_frac_phase_comp(self):
        self.mass_frac_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=0.1,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=units.dimensionless,
            doc="Mass fraction",
        )

        def rule_mass_frac_phase_comp(b, p, j):
            return (
                b.mass_frac_phase_comp[p, j] * b.flow_mass_phase[p]
                == b.flow_mass_phase_comp[p, j]
            )

        self.eq_mass_frac_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_mass_frac_phase_comp,
        )

    def _flow_vol_phase_comp(self):
        self.flow_vol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=1e-6,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.m**3 / units.s,
            doc="Component volumetric flowrate",
        )

        def rule_flow_vol_phase_comp(b, p, j):
            return (
                b.flow_vol_phase_comp[p, j] * b.dens_mass_phase_comp[p, j]
                == b.flow_mass_phase_comp[p, j]
            )

        self.eq_flow_vol_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_flow_vol_phase_comp,
        )

    def _flow_vol_phase(self):
        self.flow_vol_phase = Var(
            self.params.phase_list,
            initialize=1e-3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.m**3 / units.s,
            doc="Phase volumetric flow rate",
        )

        def rule_flow_vol_phase(b, p):
            return b.flow_vol_phase[p] == sum(
                b.flow_vol_phase_comp[p, j] for j in b.params.component_list
            )

        self.eq_flow_vol_phase = Constraint(
            self.params.phase_list, rule=rule_flow_vol_phase
        )

    def _flow_vol(self):
        self.flow_vol = Var(
            initialize=1e-3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.m**3 / units.s,
            doc="Total volumetric flow rate",
        )

        def rule_flow_vol(b):
            return b.flow_vol == sum(b.flow_vol_phase[p] for p in b.params.phase_list)

        self.eq_flow_vol = Constraint(rule=rule_flow_vol)

    def _flow_mass_phase(self):
        self.flow_mass_phase = Var(
            self.params.phase_list,
            initialize=1e-3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.kg / units.s,
            doc="Mass flow rate",
        )

        def rule_flow_mass_phase(b, p):
            return b.flow_mass_phase[p] == sum(
                b.flow_mass_phase_comp[p, j] for j in b.params.component_list
            )

        self.eq_flow_mass_phase = Constraint(
            self.params.phase_list, rule=rule_flow_mass_phase
        )

    def _dens_mass_phase(self):
        self.dens_mass_phase = Var(
            self.params.phase_list,
            initialize=1e3,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.kg * units.m**-3,
            doc="Mass density",
        )

        def rule_dens_mass_phase(b, p):
            return b.dens_mass_phase[p] * b.flow_vol_phase[p] == b.flow_mass_phase[p]

        self.eq_dens_mass_phase = Constraint(
            self.params.phase_list, rule=rule_dens_mass_phase
        )

    def _vol_frac_phase_comp(self):
        self.vol_frac_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=0.1,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=units.dimensionless,
            doc="Volumetric fraction",
        )

        def rule_vol_frac_phase_comp(b, p, j):
            return (
                b.vol_frac_phase_comp[p, j] * b.flow_vol_phase[p]
                == b.flow_vol_phase_comp[p, j]
            )

        self.eq_vol_frac_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_vol_frac_phase_comp,
        )

    def _conc_mass_phase_comp(self):
        self.conc_mass_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units.kg / units.m**3,
            doc="Mass concentration",
        )

        def rule_conc_mass_phase_comp(b, p, j):
            return (
                b.conc_mass_phase_comp[p, j]
                == b.mass_frac_phase_comp[p, j] * b.dens_mass_phase[p]
            )

        self.eq_conc_mass_phase_comp = Constraint(
            self.params.phase_list,
            self.params.component_list,
            rule=rule_conc_mass_phase_comp,
        )

    # -----------------------------------------------------------------------------
    # General Methods
    # NOTE: For scaling in the control volume to work properly, these methods must
    # return a pyomo Var or Expression

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        return self.flow_mass_phase_comp[p, j]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.none

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def define_state_vars(self):
        """Define state vars."""
        return {
            "flow_mass_phase_comp": self.flow_mass_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    # -----------------------------------------------------------------------------
    # Scaling methods
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables

        # default scaling factors have already been set with
        # idaes.core.property_base.calculate_scaling_factors()
        sf_dens_mass_phase = {}
        for p in self.params.phase_list:
            if self.is_property_constructed("dens_mass_phase"):
                sf_dens_mass_phase[p] = iscale.get_scaling_factor(
                    self.dens_mass_phase[p]
                )
            else:
                sf_dens_mass_phase[p] = self.params.get_default_scaling(
                    "dens_mass_phase", index=p
                )

        # these variables should have user input
        for p in self.params.phase_list:
            if (
                iscale.get_scaling_factor(
                    self.flow_mass_phase_comp[p, "H2O"], warning=True
                )
                is None
            ):
                iscale.set_scaling_factor(self.flow_mass_phase_comp[p, "H2O"], 1)

            if (
                iscale.get_scaling_factor(
                    self.flow_mass_phase_comp[p, "oil"], warning=True
                )
                is None
            ):
                iscale.set_scaling_factor(self.flow_mass_phase_comp[p, "oil"], 1e1)

        sf_flow_mass_phase = {}
        for p in self.params.phase_list:
            sf_flow_mass_phase[p] = 1 / (
                1 / iscale.get_scaling_factor(self.flow_mass_phase_comp[p, "H2O"])
                + 1 / iscale.get_scaling_factor(self.flow_mass_phase_comp[p, "oil"])
            )

        # these variables do not typically require user input
        if self.is_property_constructed("flow_mass_phase"):
            for p in self.params.phase_list:
                iscale.set_scaling_factor(
                    self.flow_mass_phase[p], sf_flow_mass_phase[p]
                )

        if self.is_property_constructed("mass_frac_phase_comp"):
            for p in self.params.phase_list:
                for j in self.params.component_list:
                    sf = (
                        iscale.get_scaling_factor(self.flow_mass_phase_comp[p, j])
                        / sf_flow_mass_phase[p]
                    )
                    iscale.set_scaling_factor(self.mass_frac_phase_comp[p, j], sf)

        if self.is_property_constructed("flow_vol_phase_comp"):
            for p in self.params.phase_list:
                for j in self.params.component_list:
                    sf = (
                        iscale.get_scaling_factor(self.flow_mass_phase_comp[p, j])
                        / sf_dens_mass_phase[p]
                    )
                    iscale.set_scaling_factor(self.flow_vol_phase_comp[p, j], sf)

        if self.is_property_constructed("flow_vol_phase"):
            for p in self.params.phase_list:
                sf = sf_flow_mass_phase[p] / sf_dens_mass_phase[p]
                iscale.set_scaling_factor(self.flow_vol_phase[p], sf)

        if self.is_property_constructed("flow_vol"):
            sf = sum(
                sf_flow_mass_phase[p] / sf_dens_mass_phase[p]
                for p in self.params.phase_list
            )
            iscale.set_scaling_factor(self.flow_vol, sf)

        if self.is_property_constructed("vol_frac_phase_comp"):
            for p in self.params.phase_list:
                for j in self.params.component_list:
                    sf = iscale.get_scaling_factor(
                        self.flow_vol_phase_comp[p, j]
                    ) / iscale.get_scaling_factor(self.flow_vol_phase[p])
                    iscale.set_scaling_factor(self.vol_frac_phase_comp[p, j], sf)

        if self.is_property_constructed("conc_mass_phase_comp"):
            for p in self.params.phase_list:
                for j in self.params.component_list:
                    sf = (
                        iscale.get_scaling_factor(self.mass_frac_phase_comp[p, j])
                        * sf_dens_mass_phase[p]
                    )
                    iscale.set_scaling_factor(self.conc_mass_phase_comp[p, j], sf)

        # transform property constraints
        transform_property_constraints(self)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of an ion exchange unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import Reference, units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("IonExchangeZO")
class IonExchangeZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an Ion exchange unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "ion_exchange"

        build_sido(self)
        self._Q = Reference(self.properties_in[:].flow_vol)
        pump_electricity(self, self._Q)

        # mutable parameter; default value found in WT3 for anion exchange
        if self.config.process_subtype == "clinoptilolite":
            pass
        else:
            self.eta_pump.set_value(0.8)
            # mutable parameter; default value of 2 bar converted to feet head
            self.lift_height.set_value(69.91052 * pyunits.feet)

        # Add variables and constraints for material requirements
        self.NaCl_flowrate = Var(
            self.flowsheet().time,
            initialize=1,
            units=pyunits.kg / pyunits.s,
            bounds=(0, None),
            doc="Flowrate of NaCl addition",
        )
        self.NaCl_dose = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Dosage of NaCl addition",
        )

        self._fixed_perf_vars.append(self.NaCl_dose)
        self._perf_var_dict["NaCl Addition"] = self.NaCl_flowrate

        @self.Constraint(self.flowsheet().time)
        def NaCl_constraint(blk, t):
            return blk.NaCl_flowrate[t] == blk.NaCl_dose * blk.properties_in[t].flow_vol

        self.resin_demand = Var(
            self.flowsheet().time,
            initialize=1,
            units=pyunits.kg / pyunits.s,
            bounds=(0, None),
            doc="Replacement rate of ion exchange resin",
        )
        self.resin_replacement = Var(
            units=pyunits.kg / pyunits.m**3,
            bounds=(0, None),
            doc="Resin replacement as a function of flow",
        )

        self._fixed_perf_vars.append(self.resin_replacement)
        self._perf_var_dict["Resin Demand"] = self.resin_demand

        @self.Constraint(self.flowsheet().time)
        def resin_constraint(blk, t):
            return (
                blk.resin_demand[t]
                == blk.resin_replacement * blk.properties_in[t].flow_vol
            )

        if self.config.process_subtype == "clinoptilolite":
            if "ammonium_as_nitrogen" in self.config.property_package.solute_set:
                self.nitrogen_clay_ratio = Var(
                    self.flowsheet().config.time,
                    units=pyunits.dimensionless,
                    doc="Mass fraction of nitrogen in clay mixture",
                )

                self._fixed_perf_vars.append(self.nitrogen_clay_ratio)

                self.final_solids_mass = Var(
                    self.flowsheet().config.time,
                    units=pyunits.kg / pyunits.s,
                    doc="Solids mass flow in byproduct stream",
                )

                @self.Constraint(
                    self.flowsheet().time,
                    doc="Solids mass flow in byproduct stream constraint",
                )
                def solids_mass_flow_constraint(b, t):
                    return (
                        b.final_solids_mass[t]
                        == b.properties_byproduct[t].flow_mass_comp[
                            "ammonium_as_nitrogen"
                        ]
                        / b.nitrogen_clay_ratio[t]
                    )

                self._perf_var_dict[
                    "Nitrogen-Clay Mixture Ratio (kg/kg)"
                ] = self.nitrogen_clay_ratio
                self._perf_var_dict[
                    "Final mass flow of clay and nitrogen (kg/s)"
                ] = self.final_solids_mass
            else:
                raise KeyError(
                    "ammonium_as_nitrogen should be defined in solute_list for this subtype."
                )

    @property
    def default_costing_method(self):
        return self.cost_ion_exchange

    @staticmethod
    def cost_ion_exchange(blk):
        """
        Two methods for costing ion exchange:
        (1) General method for costing ion exchange units. Capital cost is based on
        the both inlet flow and TDS.
        This method also registers the NaCl demand, resin replacement and
        electricity demand as costed flows.
        (2) General method using unit capex and unit opex cost parameters, tailored for AMO
        wastewater resource recovery (process subtype: clinoptilolite)
        """
        t0 = blk.flowsheet().time.first()

        if blk.unit_model.config.process_subtype != "clinoptilolite":
            # Get parameter dict from database
            parameter_dict = (
                blk.unit_model.config.database.get_unit_operation_parameters(
                    blk.unit_model._tech_type,
                    subtype=blk.unit_model.config.process_subtype,
                )
            )
            # Get costing parameter sub-block for this technology
            A, B, C, D = blk.unit_model._get_tech_parameters(
                blk,
                parameter_dict,
                blk.unit_model.config.process_subtype,
                [
                    "capital_a_parameter",
                    "capital_b_parameter",
                    "capital_c_parameter",
                    "capital_d_parameter",
                ],
            )

            # Add cost variable and constraint
            blk.capital_cost = pyo.Var(
                initialize=1,
                units=blk.config.flowsheet_costing_block.base_currency,
                bounds=(0, None),
                doc="Capital cost of unit operation",
            )

            ln_Q = pyo.log(
                pyo.units.convert(
                    blk.unit_model.properties_in[t0].flow_vol
                    / (pyo.units.m**3 / pyo.units.hour),
                    to_units=pyo.units.dimensionless,
                )
            )
            T = pyo.units.convert(
                blk.unit_model.properties_in[t0].conc_mass_comp["tds"]
                / (pyo.units.mg / pyo.units.liter),
                to_units=pyo.units.dimensionless,
            )

            expr = pyo.units.convert(
                pyo.exp(A + B * ln_Q + C * T + D * ln_Q * T) * pyo.units.USD_2017,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )

            blk.costing_package.add_cost_factor(
                blk, parameter_dict["capital_cost"]["cost_factor"]
            )

            blk.capital_cost_constraint = pyo.Constraint(
                expr=blk.capital_cost == blk.cost_factor * expr
            )

            # Register flows
            blk.config.flowsheet_costing_block.cost_flow(
                blk.unit_model.electricity[t0], "electricity"
            )
            blk.config.flowsheet_costing_block.cost_flow(
                blk.unit_model.NaCl_flowrate[t0], "sodium_chloride"
            )
            blk.config.flowsheet_costing_block.cost_flow(
                blk.unit_model.resin_demand[t0], "ion_exchange_resin"
            )
        else:
            # Get parameter dict from database
            parameter_dict = (
                blk.unit_model.config.database.get_unit_operation_parameters(
                    blk.unit_model._tech_type,
                    subtype=blk.unit_model.config.process_subtype,
                )
            )
            # Get costing parameter sub-block for this technology
            unit_capex, unit_opex = blk.unit_model._get_tech_parameters(
                blk,
                parameter_dict,
                blk.unit_model.config.process_subtype,
                ["unit_capex", "unit_opex"],
            )

            # Add cost variable and constraint
            blk.capital_cost = pyo.Var(
                initialize=1,
                units=blk.config.flowsheet_costing_block.base_currency,
                bounds=(0, None),
                doc="Capital cost of unit operation",
            )

            capex_expr = pyo.units.convert(
                blk.unit_model.properties_in[t0].flow_vol * unit_capex,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )

            # Determine if a costing factor is required
            blk.costing_package.add_cost_factor(
                blk, parameter_dict["capital_cost"]["cost_factor"]
            )

            blk.capital_cost_constraint = pyo.Constraint(
                expr=blk.capital_cost == blk.cost_factor * capex_expr
            )

            # Add fixed operating cost variable and constraint
            blk.fixed_operating_cost = pyo.Var(
                initialize=1,
                units=blk.config.flowsheet_costing_block.base_currency
                / blk.config.flowsheet_costing_block.base_period,
                bounds=(0, None),
                doc="Fixed operating cost of unit",
            )
            blk.fixed_operating_cost_constraint = pyo.Constraint(
                expr=blk.fixed_operating_cost
                == pyo.units.convert(
                    blk.unit_model.properties_in[t0].flow_vol * unit_opex,
                    to_units=blk.config.flowsheet_costing_block.base_currency
                    / blk.config.flowsheet_costing_block.base_period,
                )
            )

            # Register flows
            blk.config.flowsheet_costing_block.cost_flow(
                blk.unit_model.electricity[t0], "electricity"
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Test of db_api module
"""
import pytest
from ..db_api import ElectrolyteDB
from ..data_model import Component, Reaction, Base
from ..commands import _load_bootstrap
from pymongo import MongoClient
from .util import MockDB


@pytest.fixture
def mockdb():
    return MockDB()


# Test for MongoDB server at default URL
g_mongo_server = True
try:
    conn = MongoClient(
        ElectrolyteDB.DEFAULT_URL,
        serverSelectionTimeoutMS=1000,
    )
    conn.server_info()
except Exception as err:
    print(f"Cannot connect to MongoDB: {err}")
    g_mongo_server = False

requires_mongo = pytest.mark.skipif(
    g_mongo_server is False,
    reason=f"Cannot connect to MongoDB server at {ElectrolyteDB.DEFAULT_URL}",
)


@pytest.fixture(scope="module")
def edb():
    return ElectrolyteDB()


@pytest.mark.component
@requires_mongo
def test_edb_init(edb):
    assert edb is not None
    assert edb.connect_status_str == "Connection succeeded"
    assert type(edb.connect_status) is dict


@pytest.mark.component
@requires_mongo
def test_edb_load(edb):
    # Load bootstrap for temporary testing purposes
    _load_bootstrap(edb)
    base = edb.get_base("default_thermo")
    assert type(base) is Base
    edb.list_bases()


@pytest.mark.component
@requires_mongo
def test_edb_get_components(edb):
    res_obj_comps = edb.get_components(element_names=["H", "O"])
    for comp_obj in res_obj_comps:
        assert type(comp_obj) is Component

    # Just test the _process_species function
    assert edb._process_species("H2O") == "H2O"
    assert edb._process_species("H_+") == "H"
    assert edb._process_species("OH_-") == "OH"

    # Drop the bootstrap database for cleaning
    edb.drop_database(edb.DEFAULT_URL, edb.DEFAULT_DB)
    assert edb.is_empty()


@pytest.mark.unit
def test_edb_load_convention():
    # make sure the convention of collection names mapping to data_model objects is true, since
    # we rely on it in the ElectrolyteDB.load() method
    for wrapper_class in (Component, Reaction, Base):
        assert wrapper_class.__name__.lower() in ElectrolyteDB._known_collections


def insert_reactions(collection, data):
    for obj in data:
        collection.insert_one(obj)


# Data for get_reactions tests

data1 = [
    {
        "stoichiometry": {"Liq": {"H2O": -1, "CO2": -1, "H2CO3": 1}},
        "heat_of_reaction": "constant_dh_rxn",
        "equilibrium_constant": "van_t_hoff",
        "equilibrium_form": "log_power_law",
        "concentration_form": "ConcentrationForm.molarity",
        "parameter_data": {
            "dh_rxn_ref": [{"v": 0, "u": "kJ/mol", "i": 0}],
            "k_eq_ref": [{"v": 0.0017, "u": "m**3/mol", "i": 0}],
            "T_eq_ref": [{"v": 300, "u": "K", "i": 0}],
        },
        "type": "equilibrium",
        "name": "CO2_to_H2CO3",
        "components": ["CO2", "H2CO3"],
        "reactant_elements": ["C", "O", "H"],
    },
    {
        "stoichiometry": {"Liq": {"H2O": -1, "H_+": 1, "OH_-": 1}},
        "heat_of_reaction": "constant_dh_rxn",
        "equilibrium_constant": "van_t_hoff_aqueous",
        "equilibrium_form": "log_power_law",
        "concentration_form": "ConcentrationForm.molarity",
        "parameter_data": {
            "dh_rxn_ref": [{"v": 55.83, "u": "kJ/mol", "i": 0}],
            "ds_rxn_ref": [{"v": -80.7, "u": "J/mol/K", "i": 0}],
        },
        "type": "equilibrium",
        "name": "H2O_Kw",
        "components": ["H2O", "Kw"],
        "reactant_elements": ["O", "H"],
    },
]

# add one more record to data1
data2 = data1.copy() + [
    {
        "stoichiometry": {"Liq": {"H2CO3": -1, "H_+": 1, "HCO3_-": 1}},
        "heat_of_reaction": "constant_dh_rxn",
        "equilibrium_constant": "van_t_hoff_aqueous",
        "equilibrium_form": "log_power_law",
        "concentration_form": "ConcentrationForm.molarity",
        "parameter_data": {
            "dh_rxn_ref": [{"v": 7.7, "u": "kJ/mol", "i": 0}],
            "ds_rxn_ref": [{"v": -95.8, "u": "J/mol/K", "i": 0}],
        },
        "type": "equilibrium",
        "name": "H2CO3_Ka1",
        "components": ["H2CO3", "Ka1"],
        "reactant_elements": ["C", "O", "H"],
    }
]


# The way to read this parameterized test is:
#   With these 'components' and this 'data' in the DB,
#   getting reactions with *any* component should return 'any_num' records,
#   and getting reactions with *all* components should return 'all_num' records
#   and getting reactions with *all* components and *new* components should
#   return 'new_num' records
@pytest.mark.unit
@pytest.mark.parametrize(
    "components,data,any_num,all_num,new_num",
    [
        (["H2O", "CO2", "H2CO3"], data1, 2, 1, 2),
        (["H2O", "H +", "OH -", "H2CO3", "HCO3 -"], data2, 3, 2, 3),
        (["H2CO3"], data2, 2, 0, 2),
    ],
)
def test_get_reactions(mockdb, components, data, any_num, all_num, new_num):
    insert_reactions(mockdb._db.reaction, data)
    reactions = mockdb.get_reactions(components, any_components=True)
    assert len(list(reactions)) == any_num
    reactions = mockdb.get_reactions(components, any_components=False)
    assert len(list(reactions)) == all_num
    reactions = mockdb.get_reactions(
        components, any_components=False, include_new_components=True
    )
    assert len(list(reactions)) == new_num



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a AOP addition
"""

import pyomo.environ as pyo


class AOPAdditionMixin:
    @staticmethod
    def _get_aop_capital_cost(blk, A, B):
        """
        Generate expression for capital cost due to AOP addition.
        """
        t0 = blk.flowsheet().time.first()

        chemical_flow_mass = pyo.units.convert(
            blk.unit_model.chemical_flow_mass[t0], to_units=pyo.units.lb / pyo.units.day
        )
        expr = pyo.units.convert(
            A
            * pyo.units.convert(
                chemical_flow_mass / (pyo.units.lb / pyo.units.day),
                to_units=pyo.units.dimensionless,
            )
            ** B,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        return expr



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a filter press unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("FilterPressZO")
class FilterPressZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a filter press unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "filter_press"

        build_sido(self)

        self.hours_per_day_operation = Var(
            self.flowsheet().time,
            units=pyunits.hour / pyunits.day,
            doc="Hours per day of filter press operation",
        )

        self.cycle_time = Var(
            self.flowsheet().time, units=pyunits.hours, doc="Filter press cycle time"
        )

        self.electricity_a_parameter = Var(
            self.flowsheet().time,
            units=pyunits.kWh / (pyunits.year * pyunits.ft**3),
            doc="Parameter A for electricity calculation",
        )

        self.electricity_b_parameter = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Parameter B for electricity calculation",
        )

        self._fixed_perf_vars.append(self.hours_per_day_operation)
        self._fixed_perf_vars.append(self.cycle_time)
        self._fixed_perf_vars.append(self.electricity_a_parameter)
        self._fixed_perf_vars.append(self.electricity_b_parameter)

        self.filter_press_capacity = Var(
            self.flowsheet().time,
            initialize=10,
            units=pyunits.ft**3,
            doc="Filter press capacity",
        )

        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Filter press power",
        )

        @self.Constraint(self.flowsheet().time, doc="Filter press capacity constraint")
        def fp_capacity(b, t):
            Q = b.properties_in[t].flow_vol
            return b.filter_press_capacity[t] == pyunits.convert(
                Q, to_units=pyunits.ft**3 / pyunits.day
            ) / (b.hours_per_day_operation[t] / b.cycle_time[t])

        @self.Constraint(
            self.flowsheet().time, doc="Filter press electricity constraint"
        )
        def fp_electricity(b, t):
            Q = b.properties_in[t].flow_vol
            A = pyunits.convert(
                b.electricity_a_parameter[t]
                / (pyunits.kWh / (pyunits.year * pyunits.ft**3)),
                to_units=pyunits.dimensionless,
            )
            fp_cap = pyunits.convert(
                b.filter_press_capacity[t] / pyunits.ft**3,
                to_units=pyunits.dimensionless,
            )
            return b.electricity[t] == (A * fp_cap ** b.electricity_b_parameter[t]) * (
                pyunits.kWh / pyunits.year
            ) / pyunits.convert(
                Q, to_units=pyunits.m**3 / pyunits.yr
            ) * pyunits.convert(
                Q, to_units=pyunits.m**3 / pyunits.hr
            )

        self._perf_var_dict["Filter Press Capacity (ft3)"] = self.filter_press_capacity
        self._perf_var_dict["Filter Press Power (kW)"] = self.electricity

    @property
    def default_costing_method(self):
        return self.cost_filter_press

    @staticmethod
    def cost_filter_press(blk):
        """
        General method for costing belt filter press. Capital cost is a function
        of flow in gal/hr.
        """
        t0 = blk.flowsheet().time.first()
        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        Q = pyo.units.convert(
            blk.unit_model.properties_in[t0].flow_vol,
            to_units=pyo.units.gal / pyo.units.hr,
        )

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            ["capital_a_parameter", "capital_b_parameter"],
        )

        # Determine if a costing factor is required
        factor = parameter_dict["capital_cost"]["cost_factor"]

        expr = pyo.units.convert(
            A * Q + B, to_units=blk.config.flowsheet_costing_block.base_currency
        )

        blk.capital_cost_constraint = pyo.Constraint(expr=blk.capital_cost == expr)

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import (
    ConcreteModel,
    value,
    TransformationFactory,
    units as pyunits,
)
import idaes.logger as idaeslog
from pyomo.network import Arc, SequentialDecomposition
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.models.unit_models import Product
import idaes.core.util.scaling as iscale
from idaes.core import UnitModelCostingBlock

from watertap.core.util.initialization import assert_degrees_of_freedom, check_solve

from watertap.core.wt_database import Database
import watertap.core.zero_order_properties as prop_ZO
from watertap.unit_models.zero_order import (
    FeedZO,
    WaterPumpingStationZO,
    CoagulationFlocculationZO,
    SedimentationZO,
    OzoneZO,
    FixedBedZO,
    GACZO,
    UVZO,
    IonExchangeZO,
    ChlorinationZO,
    StorageTankZO,
)
from watertap.costing.zero_order_costing import ZeroOrderCosting

# Set up logger
_log = idaeslog.getLogger(__name__)


def main():
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)

    initialize_system(m)  # initialization needed for ozone unit

    results = solve(m, checkpoint="solve flowsheet after initializing system")
    display_results(m)

    add_costing(m)
    initialize_costing(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    results = solve(m, checkpoint="solve flowsheet with costing")
    display_costing(m)
    return m, results


def build():
    # flowsheet set up
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop = prop_ZO.WaterParameterBlock(solute_list=["tds", "tss", "toc"])

    # unit models
    m.fs.feed = FeedZO(property_package=m.fs.prop)
    m.fs.intake_pump = WaterPumpingStationZO(
        property_package=m.fs.prop, database=m.db, process_subtype="raw"
    )
    m.fs.coag_and_floc = CoagulationFlocculationZO(
        property_package=m.fs.prop, database=m.db
    )
    m.fs.sedimentation = SedimentationZO(property_package=m.fs.prop, database=m.db)
    m.fs.ozonation = OzoneZO(property_package=m.fs.prop, database=m.db)
    m.fs.gravity_basin = FixedBedZO(
        property_package=m.fs.prop, database=m.db, process_subtype="gravity_basin"
    )
    m.fs.gac = GACZO(
        property_package=m.fs.prop, database=m.db, process_subtype="pressure_vessel"
    )
    m.fs.backwash_pump = WaterPumpingStationZO(
        property_package=m.fs.prop, database=m.db, process_subtype="treated"
    )
    m.fs.uv = UVZO(property_package=m.fs.prop, database=m.db)
    m.fs.anion_exchange = IonExchangeZO(
        property_package=m.fs.prop, database=m.db, process_subtype="anion_exchange"
    )
    m.fs.chlorination = ChlorinationZO(property_package=m.fs.prop, database=m.db)
    m.fs.storage = StorageTankZO(property_package=m.fs.prop, database=m.db)
    m.fs.recharge_pump = WaterPumpingStationZO(
        property_package=m.fs.prop, database=m.db, process_subtype="treated"
    )
    m.fs.product = Product(property_package=m.fs.prop)

    # connections
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.intake_pump.inlet)
    m.fs.s02 = Arc(source=m.fs.intake_pump.outlet, destination=m.fs.coag_and_floc.inlet)
    m.fs.s03 = Arc(
        source=m.fs.coag_and_floc.outlet, destination=m.fs.sedimentation.inlet
    )
    m.fs.s04 = Arc(source=m.fs.sedimentation.treated, destination=m.fs.ozonation.inlet)
    m.fs.s05 = Arc(source=m.fs.ozonation.treated, destination=m.fs.gravity_basin.inlet)
    m.fs.s06 = Arc(source=m.fs.gravity_basin.treated, destination=m.fs.gac.inlet)
    m.fs.s07 = Arc(source=m.fs.gac.treated, destination=m.fs.uv.inlet)
    m.fs.s08 = Arc(source=m.fs.gac.byproduct, destination=m.fs.backwash_pump.inlet)
    m.fs.s09 = Arc(source=m.fs.uv.treated, destination=m.fs.anion_exchange.inlet)
    m.fs.s10 = Arc(
        source=m.fs.anion_exchange.treated, destination=m.fs.chlorination.inlet
    )
    m.fs.s11 = Arc(source=m.fs.chlorination.treated, destination=m.fs.storage.inlet)
    m.fs.s12 = Arc(source=m.fs.storage.outlet, destination=m.fs.recharge_pump.inlet)
    m.fs.s13 = Arc(source=m.fs.recharge_pump.outlet, destination=m.fs.product.inlet)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(m):
    # ---specifications---
    # feed
    flow_vol = 0.9224 * pyunits.m**3 / pyunits.s
    conc_mass_tds = 0.63 * pyunits.kg / pyunits.m**3
    conc_mass_tss = 0.006525 * pyunits.kg / pyunits.m**3
    conc_mass_toc = 0.004 * pyunits.kg / pyunits.m**3

    m.fs.feed.flow_vol[0].fix(flow_vol)
    m.fs.feed.conc_mass_comp[0, "tds"].fix(conc_mass_tds)
    m.fs.feed.conc_mass_comp[0, "tss"].fix(conc_mass_tss)
    m.fs.feed.conc_mass_comp[0, "toc"].fix(conc_mass_toc)
    solve(m.fs.feed, checkpoint="solve feed block")

    # intake pump
    m.fs.intake_pump.load_parameters_from_database()
    m.fs.intake_pump.electricity.fix(93.2)

    # coagulation and flocculation
    m.fs.coag_and_floc.load_parameters_from_database(use_default_removal=True)

    # sedimentation
    m.fs.sedimentation.load_parameters_from_database(use_default_removal=True)

    # # ozonation
    m.fs.ozonation.load_parameters_from_database(use_default_removal=True)

    # fixed bed gravity basin
    m.fs.gravity_basin.load_parameters_from_database(use_default_removal=True)

    # granular activated carbon
    m.fs.gac.load_parameters_from_database(use_default_removal=True)

    # backwash pump
    m.fs.backwash_pump.load_parameters_from_database()
    m.fs.backwash_pump.electricity.fix(37.3)

    # uv aop
    m.fs.uv.load_parameters_from_database(use_default_removal=True)
    m.fs.uv.uv_reduced_equivalent_dose.fix(200)
    m.fs.uv.uv_transmittance_in.fix(0.90)

    # anion exchange
    m.fs.anion_exchange.load_parameters_from_database(use_default_removal=True)
    m.fs.anion_exchange.removal_frac_mass_comp[0, "tds"].fix(0.9)

    # chlorination
    m.fs.chlorination.load_parameters_from_database(use_default_removal=True)

    # storage
    m.fs.storage.load_parameters_from_database(use_default_removal=True)
    m.fs.storage.storage_time.fix(6)

    # recharge pump
    m.fs.recharge_pump.load_parameters_from_database()
    m.fs.recharge_pump.electricity.fix(186.4)


def initialize_system(m):
    seq = SequentialDecomposition()
    seq.options.tear_set = []
    seq.options.iterLim = 1
    seq.run(m, lambda u: u.initialize())


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def display_results(m):
    unit_list = [
        "feed",
        "intake_pump",
        "coag_and_floc",
        "sedimentation",
        "ozonation",
        "gravity_basin",
        "gac",
        "backwash_pump",
        "uv",
        "anion_exchange",
        "chlorination",
        "storage",
        "recharge_pump",
        "product",
    ]

    for u in unit_list:
        m.fs.component(u).report()


def add_costing(m):
    m.fs.costing = ZeroOrderCosting()
    # typing aid
    costing_kwargs = {"flowsheet_costing_block": m.fs.costing}
    m.fs.intake_pump.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.coag_and_floc.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.sedimentation.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.ozonation.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.gravity_basin.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.gac.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.backwash_pump.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.uv.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.anion_exchange.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.chlorination.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.storage.costing = UnitModelCostingBlock(**costing_kwargs)
    m.fs.recharge_pump.costing = UnitModelCostingBlock(**costing_kwargs)

    m.fs.costing.cost_process()
    m.fs.costing.add_electricity_intensity(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)


def initialize_costing(m):
    m.fs.costing.initialize()


def display_costing(m):

    m.fs.costing.total_capital_cost.display()
    m.fs.costing.total_operating_cost.display()
    m.fs.costing.LCOW.display()

    print("\nUnit Capital Costs\n")
    for u in m.fs.costing._registered_unit_costing:
        print(
            u.name,
            " :   ",
            value(pyunits.convert(u.capital_cost, to_units=pyunits.USD_2018)),
        )

    print("\nUtility Costs\n")
    for f in m.fs.costing.used_flows:
        print(
            f,
            " :   ",
            value(
                pyunits.convert(
                    m.fs.costing.aggregate_flow_costs[f],
                    to_units=pyunits.USD_2018 / pyunits.year,
                )
            ),
        )

    print("")
    total_capital_cost = value(
        pyunits.convert(m.fs.costing.total_capital_cost, to_units=pyunits.MUSD_2018)
    )
    print(f"Total Capital Costs: {total_capital_cost:.2f} M$")
    total_operating_cost = value(
        pyunits.convert(
            m.fs.costing.total_operating_cost, to_units=pyunits.MUSD_2018 / pyunits.year
        )
    )
    print(f"Total Operating Costs: {total_operating_cost:.2f} M$/year")
    electricity_intensity = value(
        pyunits.convert(
            m.fs.costing.electricity_intensity, to_units=pyunits.kWh / pyunits.m**3
        )
    )
    print(f"Electricity Intensity: {electricity_intensity:.4f} kWh/m^3")
    LCOW = value(
        pyunits.convert(m.fs.costing.LCOW, to_units=pyunits.USD_2018 / pyunits.m**3)
    )
    print(f"Levelized Cost of Water: {LCOW:.4f} $/m^3")


if __name__ == "__main__":
    m, results = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    units as pyunits,
    Objective,
    Var,
    Constraint,
    NonNegativeReals,
    assert_optimal_termination,
    value,
)


from pyomo.network import Arc
from idaes.core import (
    FlowsheetBlock,
)

from idaes.core.solvers import get_solver
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import (
    Product,
    Feed,
    StateJunction,
)

from watertap.unit_models.nanofiltration_DSPMDE_0D import (
    NanofiltrationDSPMDE0D,
)

from watertap.unit_models.pressure_changer import Pump

from pyomo.environ import ConcreteModel, TransformationFactory

import math

from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    ActivityCoefficientModel,
    DensityCalculation,
)
import idaes.core.util.scaling as iscale

from idaes.core import UnitModelCostingBlock
from watertap.costing import WaterTAPCosting

__author__ = "Alexander Dudchenko, Adam Atia"


def main():
    solver = get_solver()
    m = build()
    initialize(m, solver)
    add_objective(m)
    unfix_opt_vars(m)
    results = optimize(m, solver)
    assert_optimal_termination(results)
    print("Optimal cost", value(m.fs.costing.LCOW))
    print("Optimal NF pressure (Bar)", m.fs.NF.pump.outlet.pressure[0].value / 1e5)
    print("Optimal area (m2)", m.fs.NF.nfUnit.area.value)
    print(
        "Optimal nf recovery (%)",
        m.fs.NF.nfUnit.recovery_vol_phase[0.0, "Liq"].value * 100,
    )

    print("Feed hardness (mg/L as CaCO3)", m.fs.feed.properties[0].total_hardness.value)
    print(
        "Product hardness (mg/L as CaCO3)",
        m.fs.product.properties[0].total_hardness.value,
    )
    print(
        "Disposal hardness (mg/L as CaCO3)",
        m.fs.disposal.properties[0].total_hardness.value,
    )
    return m


def set_default_feed(m, solver):
    """Set a default feed composition for use in flowsheet initialization routine"""
    conc_mass_phase_comp = {
        "Ca_2+": 258 / 1000,
        "HCO3_-": 385 / 1000,
        "SO4_2-": 1011 / 1000,
        "Cl_-": 870.0 / 1000,
        "Na_+": 739 / 1000,
        "K_+": 9 / 1000,
        "Mg_2+": 90 / 1000,
    }
    set_NF_feed(
        m.fs,
        solver,
        flow_mass_h2o=1,
        conc_mass_phase_comp=conc_mass_phase_comp,
        feed_mass_frac=None,
        mole_frac=None,
    )


def define_feed_comp():
    default = {
        "solute_list": [
            "Ca_2+",
            "SO4_2-",
            "HCO3_-",
            "Na_+",
            "Cl_-",
            "K_+",
            "Mg_2+",
        ],
        "diffusivity_data": {
            ("Liq", "Ca_2+"): 9.2e-10,
            ("Liq", "SO4_2-"): 1.06e-9,
            ("Liq", "HCO3_-"): 1.19e-9,
            ("Liq", "Na_+"): 1.33e-9,
            ("Liq", "Cl_-"): 2.03e-9,
            ("Liq", "K_+"): 1.957e-9,
            ("Liq", "Mg_2+"): 0.706e-9,
        },
        "mw_data": {
            "H2O": 18e-3,
            "Ca_2+": 40e-3,
            "HCO3_-": 61.0168e-3,
            "SO4_2-": 96e-3,
            "Na_+": 23e-3,
            "Cl_-": 35e-3,
            "K_+": 22.989769e-3,
            "Mg_2+": 24.305e-3,
        },
        "stokes_radius_data": {
            "Ca_2+": 0.309e-9,
            "HCO3_-": 2.06e-10,
            "SO4_2-": 0.230e-9,
            "Cl_-": 0.121e-9,
            "Na_+": 0.184e-9,
            "K_+": 0.125e-9,
            "Mg_2+": 0.347e-9,
        },
        "charge": {
            "Ca_2+": 2,
            "HCO3_-": -1,
            "SO4_2-": -2,
            "Na_+": 1,
            "Cl_-": -1,
            "K_+": 1,
            "Mg_2+": 2,
        },
        "activity_coefficient_model": ActivityCoefficientModel.ideal,
        "density_calculation": DensityCalculation.constant,
    }
    return default


def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.costing = WaterTAPCosting()
    default = define_feed_comp()
    m.fs.properties = MCASParameterBlock(**default)
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.feed.properties[0].conc_mass_phase_comp[...]
    m.fs.feed.properties[0].flow_mass_phase_comp[...]

    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)
    m.fs.product.properties[0].total_hardness

    add_hardness_constraint(m.fs.product)
    m.fs.NF = FlowsheetBlock(dynamic=False)
    build_nf_block(m, m.fs.NF)
    m.fs.feed_to_nfUnit_feed = Arc(
        source=m.fs.feed.outlet, destination=m.fs.NF.feed.inlet
    )
    m.fs.nfUnit_retentate_to_disposal = Arc(
        source=m.fs.NF.retentate.outlet,
        destination=m.fs.disposal.inlet,
    )
    m.fs.nfUnit_product_to_product = Arc(
        source=m.fs.NF.product.outlet,
        destination=m.fs.product.inlet,
    )
    m.fs.costing.disposal_cost = Var(
        initialize=0.1,
        bounds=(0, None),
        doc="disposal cost",
        units=pyunits.USD_2020 / pyunits.m**3,
    )
    m.fs.costing.disposal_cost.fix()
    m.fs.costing.register_flow_type("disposal cost", m.fs.costing.disposal_cost)
    m.fs.costing.cost_flow(
        pyunits.convert(
            m.fs.disposal.properties[0].flow_vol_phase["Liq"],
            pyunits.m**3 / pyunits.s,
        ),
        "disposal cost",
    )
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.product.properties[0].flow_vol)

    TransformationFactory("network.expand_arcs").apply_to(m)
    return m


def build_nf_block(m, blk):
    # setting up state junction
    blk.feed = StateJunction(property_package=m.fs.properties)
    blk.product = StateJunction(property_package=m.fs.properties)
    blk.retentate = StateJunction(property_package=m.fs.properties)

    blk.pump = Pump(property_package=m.fs.properties)
    blk.pump.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    # nfunit
    blk.nfUnit = NanofiltrationDSPMDE0D(property_package=m.fs.properties)
    blk.nfUnit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    blk.feed_to_pump = Arc(source=blk.feed.outlet, destination=blk.pump.inlet)
    blk.pump_to_nf = Arc(source=blk.pump.outlet, destination=blk.nfUnit.inlet)
    blk.nf_to_product = Arc(source=blk.nfUnit.permeate, destination=blk.product.inlet)
    blk.nf_to_retentate = Arc(
        source=blk.nfUnit.retentate, destination=blk.retentate.inlet
    )
    blk.nf_flux = Var(initialize=1, units=pyunits.dimensionless)
    blk.nf_flux_eq = Constraint(
        expr=blk.nf_flux == blk.nfUnit.flux_vol_water_avg[0.0] * 3600 * 1000
    )


def fix_init_vars(m):
    m.fs.feed.properties[0].pressure.fix(101325)
    m.fs.feed.properties[0].temperature.fix(293.15)

    # pump operation
    m.fs.NF.pump.outlet.pressure[0].fix(3.0 * 1e5)
    m.fs.NF.pump.efficiency_pump[0].fix(0.75)
    iscale.set_scaling_factor(m.fs.NF.pump.control_volume.work, 1e-4)
    # NF unit operation init values
    m.fs.NF.nfUnit.recovery_vol_phase[0.0, "Liq"].setub(0.95)
    m.fs.NF.nfUnit.area.fix(50)
    m.fs.NF.nfUnit.spacer_porosity.fix(0.85)
    m.fs.NF.nfUnit.spacer_mixing_efficiency.fix()
    m.fs.NF.nfUnit.spacer_mixing_length.fix()
    m.fs.NF.nfUnit.channel_height.fix(1e-3)
    m.fs.NF.nfUnit.velocity[0, 0].fix(0.25)
    m.fs.NF.nfUnit.mixed_permeate[0].pressure.fix(101325)
    # NF membrane props for NF270
    m.fs.NF.nfUnit.radius_pore.fix(0.5e-9)
    m.fs.NF.nfUnit.membrane_thickness_effective.fix(8.598945196055952e-07)
    m.fs.NF.nfUnit.membrane_charge_density.fix(-50)
    m.fs.NF.nfUnit.dielectric_constant_pore.fix(41.3)
    iscale.calculate_scaling_factors(m)


def unfix_opt_vars(m):
    m.fs.feed.properties[0].flow_vol_phase["Liq"].fix()
    for (phase, ion), obj in m.fs.feed.properties[0].conc_mass_phase_comp.items():
        m.fs.feed.properties[0].conc_mass_phase_comp["Liq", ion].fix()
        m.fs.feed.properties[0].flow_mol_phase_comp["Liq", ion].unfix()
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "H2O"].unfix()
    m.fs.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"].unfix()
    m.fs.NF.pump.outlet.pressure[0].unfix()
    m.fs.NF.nfUnit.area.unfix()
    m.fs.NF.nfUnit.velocity.unfix()
    m.fs.NF.nfUnit.velocity.setub(0.25)
    m.fs.product.max_hardness.fix(500)
    m.fs.NF.nfUnit.recovery_vol_phase[0.0, "Liq"].setub(0.9)
    # Touch total_hardness (on-demand property) at feed and disposal for reporting
    m.fs.feed.properties[0].total_hardness
    m.fs.disposal.properties[0].total_hardness


def add_objective(m):
    if m.find_component("fs.LCOW_objective") is None:
        m.fs.LCOW_objective = Objective(expr=m.fs.costing.LCOW)
        print("added objective function")


def optimize(m, solver=None, **kwargs):
    if solver is None:
        solver = get_solver()
    # add_objective(m)
    print("Optimizing with {} DOFs".format(degrees_of_freedom(m)))
    result = solver.solve(m, tee=True)
    return result


def initialize(m, solver=None, **kwargs):
    if solver is None:
        solver = get_solver()
    set_default_feed(m, solver)
    fix_init_vars(m)
    init_system(m, solver)
    # solve box problem
    print("initalized, DOFs:", degrees_of_freedom(m))
    assert degrees_of_freedom(m) == 0
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)
    print("Solved box problem")


def init_system(m, solver):
    if solver is None:
        solver = get_solver()
    m.fs.feed.initialize(optarg=solver.options)

    propagate_state(m.fs.feed_to_nfUnit_feed)

    init_nf_block(m.fs.NF, solver)

    propagate_state(m.fs.nfUnit_retentate_to_disposal)
    propagate_state(m.fs.nfUnit_product_to_product)
    m.fs.NF.product.initialize(optarg=solver.options)
    m.fs.NF.retentate.initialize(optarg=solver.options)

    m.fs.costing.initialize()


def init_nf_block(blk, solver):
    if solver is None:
        solver = get_solver()
    blk.feed.initialize(optarg=solver.options)
    propagate_state(blk.feed_to_pump)
    blk.pump.initialize(optarg=solver.options)
    propagate_state(blk.pump_to_nf)
    blk.nfUnit.initialize(optarg=solver.options)
    propagate_state(blk.nf_to_product)
    propagate_state(blk.nf_to_retentate)
    blk.product.initialize(optarg=solver.options)
    blk.retentate.initialize(optarg=solver.options)


def set_NF_feed(
    blk,
    solver,
    flow_mass_h2o,
    conc_mass_phase_comp=None,
    feed_mass_frac=None,
    mole_frac=None,
):
    if solver is None:
        solver = get_solver()

    mass_flow_in = flow_mass_h2o * pyunits.kg / pyunits.s
    blk.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix(flow_mass_h2o)

    if feed_mass_frac is not None:
        for ion, x in feed_mass_frac.items():
            mol_comp_flow = (
                x
                * pyunits.kg
                / pyunits.kg
                * mass_flow_in
                / blk.feed.properties[0].mw_comp[ion]
            )
            blk.feed.properties[0].flow_mol_phase_comp["Liq", ion].fix(mol_comp_flow)
        H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
        H2O_mol_comp_flow = (
            H2O_mass_frac
            * pyunits.kg
            / pyunits.kg
            * mass_flow_in
            / blk.feed.properties[0].mw_comp["H2O"]
        )
        blk.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"].fix(H2O_mol_comp_flow)

    if mole_frac is not None:
        for ion, x in mole_frac.items():
            blk.feed.properties[0].flow_mol_phase_comp["Liq", ion].fix(x)
    if conc_mass_phase_comp is not None:
        for ion, x in conc_mass_phase_comp.items():
            blk.feed.properties[0].conc_mass_phase_comp["Liq", ion].fix(x)
            blk.feed.properties[0].flow_mol_phase_comp["Liq", ion].unfix()
        solver.solve(blk.feed)
        blk.feed.properties[0].conc_mass_phase_comp["Liq", "H2O"].fix()
        for ion, x in conc_mass_phase_comp.items():
            blk.feed.properties[0].conc_mass_phase_comp["Liq", ion].unfix()
            blk.feed.properties[0].flow_mol_phase_comp["Liq", ion].fix()
            blk.feed.properties[0].flow_mass_phase_comp["Liq", ion].unfix()
        blk.feed.properties[0].conc_mass_phase_comp["Liq", "H2O"].unfix()
        blk.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].unfix()
        blk.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"].fix()
    set_NF_feed_scaling(blk)

    blk.feed.properties[0].assert_electroneutrality(
        defined_state=True,
        adjust_by_ion="Cl_-",
        get_property="flow_mol_phase_comp",
    )

    blk.feed.properties[0].temperature.fix(298.15)

    # switching to concentration for ease of adjusting in UI
    for ion, x in conc_mass_phase_comp.items():
        blk.feed.properties[0].conc_mass_phase_comp["Liq", ion].unfix()
        blk.feed.properties[0].flow_mol_phase_comp["Liq", ion].fix()


def calc_scale(value):
    return -1 * math.floor(math.log(value, 10))


def set_NF_feed_scaling(blk):
    _add = 0
    for index in blk.feed.properties[0].flow_mol_phase_comp:
        scale = calc_scale(blk.feed.properties[0].flow_mol_phase_comp[index].value)
        print(f"{index} flow_mol_phase_comp scaling factor = {10**(scale+_add)}")
        blk.properties.set_default_scaling(
            "flow_mol_phase_comp", 10 ** (scale + _add), index=index
        )


def add_hardness_constraint(stream):
    stream.max_hardness = Var(
        initialize=10000,
        domain=NonNegativeReals,
        units=pyunits.mg / pyunits.L,
        doc="Maximum total hardness as CaCO3",
    )

    stream.max_hardness_constraint = Constraint(
        expr=stream.properties[0].total_hardness <= stream.max_hardness
    )


if __name__ == "__main__":
    m = main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    value,
    Constraint,
    Objective,
    Var,
    Param,
    NonNegativeReals,
    TransformationFactory,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import (
    propagate_state,
)
from idaes.models.unit_models import Product, Feed
from idaes.core import UnitModelCostingBlock
import idaes.core.util.scaling as iscale
from idaes.core.util.misc import StrEnum

import watertap.property_models.NaCl_prop_pack as props
from watertap.unit_models.reverse_osmosis_0D import (
    ReverseOsmosis0D,
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)
from watertap.unit_models.osmotically_assisted_reverse_osmosis_0D import (
    OsmoticallyAssistedReverseOsmosis0D,
)
from watertap.unit_models.pressure_changer import Pump, EnergyRecoveryDevice
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.costing import WaterTAPCosting


class ERDtype(StrEnum):
    pump_as_turbine = "pump_as_turbine"


def erd_type_not_found(erd_type):
    raise NotImplementedError(
        "erd_type was {}, but can only " "be pump_as_turbine" "".format(erd_type.value)
    )


def main(erd_type=ERDtype.pump_as_turbine, raise_on_failure=False):
    # set up solver
    solver = get_solver()

    # build, set, and initialize
    m = build(erd_type=erd_type)
    set_operating_conditions(m)
    initialize_system(m, solver=solver)

    optimize_set_up(m)
    solve(m, solver=solver)

    print("\n***---Simulation results---***")
    display_system(m)
    if erd_type == ERDtype.pump_as_turbine:
        display_state(m)
    else:
        pass

    return m


def build(erd_type=ERDtype.pump_as_turbine):
    # TODO: add costing later (OARO unit model does not have costing method)

    # flowsheet set up
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.erd_type = erd_type
    m.fs.properties = props.NaClParameterBlock()
    m.fs.costing = WaterTAPCosting()

    # Control volume flow blocks
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    # --- Main pump ---
    m.fs.P1 = Pump(property_package=m.fs.properties)
    m.fs.P1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.P2 = Pump(property_package=m.fs.properties)
    m.fs.P2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    m.fs.P3 = Pump(property_package=m.fs.properties)
    m.fs.P3.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- Reverse Osmosis Block ---
    m.fs.RO = ReverseOsmosis0D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        has_full_reporting=True,
    )
    m.fs.RO.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- Osmotically Assisted Reverse Osmosis Block ---
    m.fs.OARO = OsmoticallyAssistedReverseOsmosis0D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        pressure_change_type=PressureChangeType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        has_full_reporting=True,
    )
    m.fs.OARO.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # --- ERD blocks ---
    if erd_type == ERDtype.pump_as_turbine:
        # add energy recovery turbine block
        m.fs.ERD1 = EnergyRecoveryDevice(property_package=m.fs.properties)
        m.fs.ERD2 = EnergyRecoveryDevice(property_package=m.fs.properties)
        # add costing for ERD config
        m.fs.ERD1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.ERD2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    else:
        erd_type_not_found(erd_type)

    # process costing and add system level metrics
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_electrical_carbon_intensity(
        m.fs.product.properties[0].flow_vol
    )

    # system water recovery
    m.fs.volumetric_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Volumetric Recovery of Water",
    )
    m.fs.eq_volumetric_recovery = Constraint(
        expr=m.fs.feed.properties[0].flow_vol_phase["Liq"] * m.fs.volumetric_recovery
        == m.fs.product.properties[0].flow_vol_phase["Liq"]
    )

    m.fs.water_recovery = Var(
        initialize=0.5,
        bounds=(0, 1),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Water Recovery",
    )
    m.fs.eq_water_recovery = Constraint(
        expr=m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"]
        * m.fs.water_recovery
        == m.fs.product.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    )

    # connections
    if erd_type == ERDtype.pump_as_turbine:
        m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.P1.inlet)
        m.fs.s02 = Arc(source=m.fs.P1.outlet, destination=m.fs.OARO.feed_inlet)
        m.fs.s03 = Arc(source=m.fs.OARO.feed_outlet, destination=m.fs.ERD1.inlet)
        m.fs.s04 = Arc(source=m.fs.ERD1.outlet, destination=m.fs.disposal.inlet)
        m.fs.s05 = Arc(source=m.fs.OARO.permeate_outlet, destination=m.fs.P2.inlet)
        m.fs.s06 = Arc(source=m.fs.P2.outlet, destination=m.fs.RO.inlet)
        m.fs.s07 = Arc(source=m.fs.RO.permeate, destination=m.fs.product.inlet)
        m.fs.s08 = Arc(source=m.fs.RO.retentate, destination=m.fs.ERD2.inlet)
        m.fs.s09 = Arc(source=m.fs.ERD2.outlet, destination=m.fs.P3.inlet)
        m.fs.s10 = Arc(source=m.fs.P3.outlet, destination=m.fs.OARO.permeate_inlet)

    else:
        # this case should be caught in the previous conditional
        erd_type_not_found(erd_type)

    TransformationFactory("network.expand_arcs").apply_to(m)

    # scaling
    # set default property values
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
    )
    # set unit model values
    iscale.set_scaling_factor(m.fs.P1.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.P2.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.P3.control_volume.work, 1e-3)
    iscale.set_scaling_factor(m.fs.RO.area, 1e-2)
    iscale.set_scaling_factor(m.fs.OARO.area, 1e-2)
    m.fs.feed.properties[0].flow_vol_phase["Liq"]
    m.fs.feed.properties[0].mass_frac_phase_comp["Liq", "NaCl"]
    if erd_type == ERDtype.pump_as_turbine:
        iscale.set_scaling_factor(m.fs.ERD1.control_volume.work, 1e-3)
        iscale.set_scaling_factor(m.fs.ERD2.control_volume.work, 1e-3)
    else:
        erd_type_not_found(erd_type)
    # unused scaling factors needed by IDAES base costing module
    # calculate and propagate scaling factors
    iscale.calculate_scaling_factors(m)

    return m


def set_operating_conditions(
    m,
    solver=None,
):

    if solver is None:
        solver = get_solver()

    # ---specifications---
    # feed
    # state variables
    pressure_atmospheric = 101325
    feed_temperature = 273.15 + 25
    m.fs.feed.properties[0].pressure.fix(pressure_atmospheric)  # feed pressure [Pa]
    m.fs.feed.properties[0].temperature.fix(feed_temperature)  # feed temperature [K]

    # properties (cannot be fixed for initialization routines, must calculate the state variables)
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    # pump 1, high pressure pump, 2 degrees of freedom (efficiency and outlet pressure)
    m.fs.P1.efficiency_pump.fix(0.80)  # pump efficiency [-]
    m.fs.P1.control_volume.properties_out[0].pressure.fix(50e5)

    # pump 2, 2 degrees of freedom (efficiency and outlet pressure)
    m.fs.P2.efficiency_pump.fix(0.80)  # pump efficiency [-]
    m.fs.P2.control_volume.properties_out[0].pressure.fix(18e5)

    # pump 3, 2 degrees of freedom (efficiency, outlet pressure)
    m.fs.P3.efficiency_pump.fix(0.80)  # pump efficiency [-]
    m.fs.P3.control_volume.properties_out[0].pressure.fix(4e5)

    # initial guess for states of pump 3 output (temperature and concentrations)
    m.fs.P3.control_volume.properties_out[0].temperature.value = feed_temperature
    permeate_flow_mass = 0.75
    permeate_mass_frac_NaCl = 0.015
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.P3.control_volume.properties_out[0].flow_mass_phase_comp[
        "Liq", "H2O"
    ].value = (permeate_flow_mass * permeate_mass_frac_H2O)
    m.fs.P3.control_volume.properties_out[0].flow_mass_phase_comp[
        "Liq", "NaCl"
    ].value = (permeate_flow_mass * permeate_mass_frac_NaCl)

    # Initialize OARO
    membrane_area = 50
    A = 4.2e-12
    B = 1.3e-8

    m.fs.OARO.area.fix(membrane_area)

    m.fs.OARO.A_comp.fix(A)
    m.fs.OARO.B_comp.fix(B)

    m.fs.OARO.structural_parameter.fix(300e-6)

    m.fs.OARO.permeate_side.channel_height.fix(0.001)
    m.fs.OARO.permeate_side.spacer_porosity.fix(0.75)
    m.fs.OARO.feed_side.channel_height.fix(0.002)
    m.fs.OARO.feed_side.spacer_porosity.fix(0.75)
    m.fs.OARO.feed_side.velocity[0, 0].fix(0.1)

    # RO unit
    m.fs.RO.A_comp.fix(4.2e-12)  # membrane water permeability coefficient [m/s-Pa]
    m.fs.RO.B_comp.fix(3.5e-8)  # membrane salt permeability coefficient [m/s]
    m.fs.RO.feed_side.channel_height.fix(1e-3)  # channel height in membrane stage [m]
    m.fs.RO.feed_side.spacer_porosity.fix(0.97)  # spacer porosity in membrane stage [-]
    m.fs.RO.permeate.pressure[0].fix(101325)  # atmospheric pressure [Pa]
    m.fs.RO.width.fix(5)  # stage width [m]
    m.fs.RO.area.fix(50)  # guess area for RO initialization

    if m.fs.erd_type == ERDtype.pump_as_turbine:
        # energy recovery turbine - efficiency and outlet pressure
        m.fs.ERD1.efficiency_pump.fix(0.95)
        m.fs.ERD1.control_volume.properties_out[0].pressure.fix(pressure_atmospheric)

        m.fs.ERD2.efficiency_pump.fix(0.95)
        m.fs.ERD2.control_volume.properties_out[0].pressure.fix(pressure_atmospheric)
    else:
        erd_type_not_found(m.fs.erd_type)

    # check degrees of freedom
    if degrees_of_freedom(m) != 0:
        raise RuntimeError(
            "The set_operating_conditions function resulted in {} "
            "degrees of freedom rather than 0. This error suggests "
            "that too many or not enough variables are fixed for a "
            "simulation.".format(degrees_of_freedom(m))
        )


def solve(blk, solver=None, tee=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    if not check_optimal_termination(results):
        results = solver.solve(blk, tee=tee)
    return results


def initialize_loop(m, solver):

    propagate_state(m.fs.s05)
    m.fs.P2.initialize()

    # ---initialize RO---
    propagate_state(m.fs.s06)
    m.fs.RO.initialize()
    propagate_state(m.fs.s07)

    propagate_state(m.fs.s08)
    m.fs.ERD2.initialize()

    propagate_state(m.fs.s09)
    m.fs.P3.initialize()

    propagate_state(m.fs.s10)
    propagate_state(m.fs.s02)
    m.fs.OARO.initialize()

    propagate_state(m.fs.s03)
    m.fs.ERD1.initialize()
    propagate_state(m.fs.s04)
    m.fs.disposal.initialize()


def initialize_system(m, solver=None, verbose=True):
    if solver is None:
        solver = get_solver()

    # ---initialize feed block---
    m.fs.feed.initialize()

    propagate_state(m.fs.s01)
    m.fs.P1.initialize()

    # ---initialize OARO---
    propagate_state(m.fs.s02)
    propagate_state(m.fs.s10)
    m.fs.OARO.initialize()

    initialize_loop(m, solver)

    print(f"DOF: {degrees_of_freedom(m)}")

    # Now that the units are initialized, we can fix the
    # permeate side outlet pressure and unfix the RO pump
    # (which allows for control over the flow mass composition
    # into the OARO permeate_side).
    m.fs.OARO.permeate_side.properties_out[0].pressure.fix(101325)
    m.fs.P2.control_volume.properties_out[0].pressure.unfix()

    print(f"DOF: {degrees_of_freedom(m)}")

    m.fs.costing.initialize()


def optimize_set_up(m):
    # add objective
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # unfix decision variables and add bounds
    # pump 1
    m.fs.P1.control_volume.properties_out[0].pressure.unfix()
    m.fs.P1.control_volume.properties_out[0].pressure.setlb(10e5)
    m.fs.P1.control_volume.properties_out[0].pressure.setub(80e5)
    m.fs.P1.deltaP.setlb(0)

    # RO
    m.fs.RO.area.unfix()
    m.fs.RO.area.setlb(1)
    m.fs.RO.area.setub(150)

    # additional specifications
    m.fs.product_salinity = Param(
        initialize=500e-6, mutable=True
    )  # product NaCl mass fraction [-]
    m.fs.minimum_water_flux = Param(
        initialize=1.0 / 3600.0, mutable=True
    )  # minimum water flux [kg/m2-s]

    # additional constraints
    m.fs.eq_product_quality = Constraint(
        expr=m.fs.product.properties[0].mass_frac_phase_comp["Liq", "NaCl"]
        <= m.fs.product_salinity
    )
    iscale.constraint_scaling_transform(
        m.fs.eq_product_quality, 1e3
    )  # scaling constraint
    m.fs.eq_minimum_water_flux = Constraint(
        expr=m.fs.RO.flux_mass_phase_comp[0, 1, "Liq", "H2O"] >= m.fs.minimum_water_flux
    )

    # ---checking model---
    assert_degrees_of_freedom(m, 2)


def optimize(m, solver=None):
    # --solve---
    return solve(m, solver=solver)


def display_system(m):
    print("---system metrics---")
    feed_flow_mass = sum(
        m.fs.feed.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    feed_mass_frac_NaCl = (
        m.fs.feed.flow_mass_phase_comp[0, "Liq", "NaCl"].value / feed_flow_mass
    )
    print("Feed: %.2f kg/s, %.0f ppm" % (feed_flow_mass, feed_mass_frac_NaCl * 1e6))

    prod_flow_mass = sum(
        m.fs.product.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
    )
    prod_mass_frac_NaCl = (
        m.fs.product.flow_mass_phase_comp[0, "Liq", "NaCl"].value / prod_flow_mass
    )
    print("Product: %.3f kg/s, %.0f ppm" % (prod_flow_mass, prod_mass_frac_NaCl * 1e6))

    print("Volumetric recovery: %.1f%%" % (value(m.fs.volumetric_recovery) * 100))
    print("Water recovery: %.1f%%" % (value(m.fs.water_recovery) * 100))

    print(
        "Energy Consumption: %.1f kWh/m3"
        % value(m.fs.costing.specific_energy_consumption)
    )
    print("Levelized cost of water: %.2f $/m3" % value(m.fs.costing.LCOW))


def display_design(m):
    print("---decision variables---")
    print("Operating pressure %.1f bar" % (m.fs.RO.inlet.pressure[0].value / 1e5))
    print("Membrane area %.1f m2" % (m.fs.RO.area.value))

    print("---design variables---")
    print(
        "Pump 1\noutlet pressure: %.1f bar\npower %.2f kW"
        % (
            m.fs.P1.outlet.pressure[0].value / 1e5,
            m.fs.P1.work_mechanical[0].value / 1e3,
        )
    )
    if m.fs.erd_type == ERDtype.pump_as_turbine:
        print(
            "ERD\ninlet pressure: %.1f bar\npower recovered %.2f kW"
            % (
                m.fs.ERD.inlet.pressure[0].value / 1e5,
                -1 * m.fs.ERD.work_mechanical[0].value / 1e3,
            )
        )
    else:
        erd_type_not_found(m.fs.erd_type)


def display_state(m):
    print("---state---")

    def print_state(s, b):
        flow_mass = sum(
            b.flow_mass_phase_comp[0, "Liq", j].value for j in ["H2O", "NaCl"]
        )
        mass_frac_ppm = b.flow_mass_phase_comp[0, "Liq", "NaCl"].value / flow_mass * 1e6
        pressure_bar = b.pressure[0].value / 1e5
        print(
            s
            + ": %.3f kg/s, %.0f ppm, %.1f bar"
            % (flow_mass, mass_frac_ppm, pressure_bar)
        )

    print_state("Feed      ", m.fs.feed.outlet)
    print_state("P1 out    ", m.fs.P1.outlet)
    print_state("OARO Feed in", m.fs.OARO.feed_inlet)
    print_state("OARO Feed out", m.fs.OARO.feed_outlet)
    print_state("OARO Perm in", m.fs.OARO.permeate_inlet)
    print_state("OARO Perm out", m.fs.OARO.permeate_outlet)
    print_state("ERD1 out    ", m.fs.ERD1.outlet)
    print_state("P2 out    ", m.fs.P2.outlet)
    print_state("RO reten  ", m.fs.RO.retentate)
    print_state("RO perm   ", m.fs.RO.permeate)
    print_state("ERD2 out    ", m.fs.ERD2.outlet)
    print_state("P3 out    ", m.fs.P3.outlet)


if __name__ == "__main__":
    m = main(erd_type=ERDtype.pump_as_turbine)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from idaes.core import declare_process_block_class, ProcessBlockData, UnitModelBlockData
from idaes.core.util.misc import add_object_reference
from idaes.core.base.costing_base import (
    UnitModelCostingBlockData,
    UnitModelCostingBlock,
    assert_flowsheet_costing_block,
    DefaultCostingComponents,
)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("MultiUnitModelCostingBlock")
class MultiUnitModelCostingBlockData(UnitModelCostingBlockData, UnitModelCostingBlock):
    """
    Class for constructing several costing blocks on the same
    unit model and then allowing for choice between them
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "flowsheet_costing_block",
        ConfigValue(
            domain=assert_flowsheet_costing_block,
            doc="Reference to associated FlowsheetCostingBlock to use.",
        ),
    )
    CONFIG.declare(
        "costing_blocks",
        ConfigValue(
            domain=dict,
            doc="Costing blocks to use for unit. Should be a dictionary whose keys "
            "are the names of the block and whose values are either the costing "
            "method used to construct the block, or another dictionary whose keys "
            "are `costing_method`, which has the value of the costing method, and "
            "optionally `costing_method_arguments`, which has the keyword arguments "
            "for the costing method specified.",
        ),
    )
    CONFIG.declare(
        "initial_costing_block",
        ConfigValue(
            doc="Costing block to be initially active",
            default=None,
        ),
    )

    def build(self):
        ProcessBlockData.build(self)

        # Alias flowsheet costing block reference
        fcb = self.config.flowsheet_costing_block

        # Get reference to unit model
        unit_model = self.parent_block()

        # Check that parent is an instance of a UnitModelBlockData
        if UnitModelBlockData not in unit_model.__class__.__mro__:
            raise TypeError(
                f"{self.name} - parent object ({unit_model.name}) is not an "
                f"instance of a UnitModelBlockData object. "
                "UnitModelCostingBlocks can only be added to UnitModelBlocks."
            )

        # Check to see if unit model already has costing
        for b in unit_model.component_objects(pyo.Block, descend_into=False):
            if b is not self and isinstance(
                b, (UnitModelCostingBlock, MultiUnitModelCostingBlock)
            ):
                # Block already has costing, clean up and raise exception
                raise RuntimeError(
                    f"Unit model {unit_model.name} already has a costing block"
                    f" registered: {b.name}. Each unit may only have a single "
                    "UnitModelCostingBlock associated with it."
                )
        # Add block to unit model initialization order
        unit_model._initialization_order.append(self)

        # Register unit model with this costing package
        fcb._registered_unit_costing.append(self)

        # Assign object references for costing package and unit model
        add_object_reference(self, "costing_package", fcb)
        add_object_reference(self, "unit_model", unit_model)

        self.costing_blocks = pyo.Block(self.config.costing_blocks)
        self.costing_block_selector = pyo.Param(
            self.config.costing_blocks, domain=pyo.Boolean, default=0, mutable=True
        )

        if self.config.initial_costing_block is None:
            for k in self.config.costing_blocks:
                self.costing_block_selector[k].set_value(1)
                break
        else:
            self.costing_block_selector[self.config.initial_costing_block].set_value(1)

        # Get costing method if not provided
        for k, val in self.config.costing_blocks.items():

            # if we get a callable, just use it
            if callable(val):
                method = val
                kwds = {}
            # else we'll assume it's a dictionary with keys
            # `costing_method` and potentially `costing_method_arguments`.
            # We raise an error if we find something else.
            else:
                for l in val:
                    if l not in ("costing_method", "costing_method_arguments"):
                        raise RuntimeError(
                            f"Unrecognized key {l} for costing block {k}."
                        )
                try:
                    method = val["costing_method"]
                except KeyError:
                    raise KeyError(
                        f"Must specify a `costing_method` key for costing "
                        f"block {k}."
                    )
                kwds = val.get("costing_method_arguments", {})

            blk = self.costing_blocks[k]

            # Assign object references for costing package and unit model
            add_object_reference(blk, "costing_package", fcb)
            add_object_reference(blk, "unit_model", unit_model)

            # Call unit costing method
            method(blk, **kwds)

            # Check that costs are Vars and have lower bound of 0
            cost_vars = DefaultCostingComponents
            for v in cost_vars:
                try:
                    cvar = getattr(blk, v)
                    if not isinstance(cvar, pyo.Var):
                        raise TypeError(
                            f"{unit_model.name} {v} component must be a Var. "
                            "Please check the costing package you are using to "
                            "ensure that all costing components are declared as "
                            "variables."
                        )
                    elif cvar.lb is None or cvar.lb < 0:
                        _log.warning(
                            f"{unit_model.name} {v} component has a lower bound "
                            "less than zero. Be aware that this may result in "
                            "negative costs during optimization."
                        )
                except AttributeError:
                    pass

        # Now we need to tie them all together
        cost_vars = list(DefaultCostingComponents) + ["direct_capital_cost"]
        for vname in cost_vars:
            for blk in self.costing_blocks.values():
                if hasattr(blk, vname):
                    break
            else:  # no break
                continue

            expr = 0.0
            for name, blk in self.costing_blocks.items():
                cvar = blk.component(vname)
                if cvar is None:
                    continue
                expr += self.costing_block_selector[name] * cvar

            self.add_component(vname, pyo.Expression(expr=expr))

    def select_costing_block(self, costing_block_name):
        """
        Set the active costing block
        """
        # zero out everything else
        self.costing_block_selector[:].set_value(0)
        self.costing_block_selector[costing_block_name].set_value(1)

    def initialize(self, *args, **kwargs):
        """
        Initialize all costing blocks for easy switching between
        """
        # TODO: Implement an initialization method
        # TODO: Need to have a general purpose method (block triangularisation?)
        # TODO: Should also allow registering custom methods

        # Vars and Constraints
        for blk in self.costing_blocks.values():
            for c in DefaultCostingComponents:
                if hasattr(blk, c):
                    var = getattr(blk, c)
                    cons = getattr(blk, f"{c}_constraint")
                    calculate_variable_from_constraint(var, cons)
